"""
SPDX-License-Identifier: GPL-3.0-or-later

Retrieve configuration for use with the rest of execution.

Read the ncmonitor(5) manpage for an overview of configuration.
"""

from logging import getLogger
import os
import re
import stat

from yaml import safe_load

DEFAULT_CONFIG_PATH = '/etc/ncmonitor/ncmonitor.yaml'
VALID_HOSTNAME_RE = r'^([a-z\d](-*[a-z\d])*)(\.([a-z\d](-*[a-z\d])*))*$'

logger = getLogger(__name__)


class NCConfigError(Exception):
    """
    Base exception class.
    """


def is_valid_hostname(hostname: str) -> bool:
    """
    Rough validatation of the configured domain name.
    """
    logger.debug("Validating hostname %s", hostname)
    compiled = re.compile(VALID_HOSTNAME_RE, re.IGNORECASE)
    is_valid = bool(compiled.match(hostname))
    logger.debug("Does %s match domain validation regex? %s", hostname, is_valid)
    return is_valid


# pylint: disable-next=too-few-public-methods
class NCMonitorConfig:
    """
    Represents a configuration for ncmonitor
    """

    def __init__(self, path: str = DEFAULT_CONFIG_PATH) -> None:
        self.config_file = {}
        self.path = path
        self.nameservers = []
        self.mark_user = None
        self.mark_pass = None

        self._load_config(self.path)

    def _load_config(self, load_path: str) -> None:
        """
        Load a configuration file from the specified filename.
        """
        logger.debug("Loading configuration file %s.", load_path)
        # We have credentials in this file
        if self._is_valid_mode():
            with open(load_path, encoding='utf-8') as conf_file:
                self.config_file = safe_load(conf_file)
        else:
            logger.exception("%s is world-readable. Exiting", self.config_file)
            raise NCConfigError

        self.nameservers.clear()
        try:
            for host in self.config_file["nameservers"]:
                if is_valid_hostname(host):
                    self.nameservers.append(host)
                else:
                    logger.exception("%s is not a valid hostname.", host)
                    raise NCConfigError
            logger.debug("Configured nameservers: %s", self.nameservers)
        except (KeyError, TypeError) as exc:
            logger.exception("Invalid config file; Have you specified any nameservers?")
            raise NCConfigError from exc

        try:
            self.mark_user = self.config_file["markmonitor"]["username"]
            self.mark_pass = self.config_file["markmonitor"]["password"]
        except (KeyError, TypeError) as exc:
            logger.exception("Missing credentials for MarkMonitor")
            raise NCConfigError from exc

    def _is_valid_mode(self) -> bool:
        """
        Validate the passed-in file's permissions mode to make sure it is not
        world-readable.
        """
        conf_stat = os.stat(self.path)
        logger.debug("Checking file attributes for %s", self.path)
        conf_mode = stat.S_IMODE(conf_stat.st_mode)
        insecure = stat.S_IMODE(conf_mode) & (stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH)
        logger.debug("* File stat is %s", conf_stat)
        logger.debug("* Raw file mode is %s", conf_mode)
        logger.debug("* File mode is %s", oct(conf_mode))
        logger.debug("* File mode considered insecure: %s", insecure)
        return not insecure
