"""
SPDX-License-Identifier: GPL-3.0-or-later
"""

from argparse import ArgumentParser
from collections import defaultdict
import json
import os
import logging
from pathlib import Path

from .config import NCMonitorConfig
from .gitrepo import DNSRepository, ACMEChiefConfig, NCRedirRedirects
from .markmonitor import MarkMonitor

logger = logging.getLogger(__name__)


class SetToListEncoder(json.JSONEncoder):
    """
    JSON encoder that handles serialization of sets.
    """
    # pylint: disable-next=arguments-renamed
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def get_conf_file() -> str:
    """
    Locate user-supplied configuration file or check in pre-configured paths.
    """
    conf_name = Path("ncmonitor.yaml")
    default_paths = [
        Path(Path.cwd() / conf_name),
        Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config" / "ncmonitor" / conf_name)),
        Path("/etc/ncmonitor" / conf_name),
    ]
    for candidate in default_paths:
        if isinstance(candidate, Path) and candidate.is_file():
            return candidate

    logger.exception("Unable to locate any configuration file.\nSearched in %s", default_paths)
    raise FileNotFoundError


def parse_args() -> ArgumentParser:
    """
    Parse runtime arguments passed in.
    """
    parser = ArgumentParser(
        description="Look up registrar DNS values and sync them with other services."
    )
    parser.add_argument("-c", "--config", type=str, help="load specified configuration file")
    parser.add_argument("-d", "--debug", action="store_true", help="Show debugging output")
    parser.add_argument("-p", "--pprint", action="store_true", help="Pretty-print JSON output")
    parser.add_argument("-v", "--verbose", action="store_true", help="Show verbose output")
    return parser.parse_args()


def setup_logging(args: ArgumentParser) -> None:
    """
    Configure the runtime according to passed arguments and the configuration
    file.
    """
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    elif args.verbose:
        logging.basicConfig(level=logging.INFO)


def get_domains(mark_user: str, mark_pass: str) -> dict:
    """
    Reach out, populate, and return domains from each service. MarkMonitor's
    entries are different from the others: Rather than just a list of domains,
    it features a list of definitions. Each domain has a value containing its
    nameservers.
    """
    domains = defaultdict(set)
    domains["mark"] = defaultdict(dict)

    with ACMEChiefConfig() as acme:
        for domain in acme.get_domain_list():
            domains["acme"].add(domain.domain_name)

    with NCRedirRedirects() as ncredir:
        for domain in ncredir.get_domain_list():
            domains["ncredir"].add(domain.domain_name)

    with MarkMonitor(mark_user, mark_pass) as mark:
        for domain in mark.get_domain_list():
            domains["mark"][domain.domain_name] = domain.nameservers

    with DNSRepository() as dnsrepo:
        for domain in dnsrepo.get_domain_list():
            domains["dnsrepo"].add(domain.domain_name)

    logger.debug("Collected list of domains from services:\n%s", domains)
    return domains


def process_domains(nameservers: list, domains: dict, pretty_print: bool = False) -> str:
    """
    Compare and output domain inconsistencies in JSON.
    """
    # MarkMonitor is our definitive list of domains; All other services are
    # compared against MarkMonitor's list. "Valid" domains in MarkMonitor are
    # therefore judged a little differently: Rather than comparing whether the
    # domains are also in some other list, we just need to verify if the
    # nameservers match what we expect.
    results = defaultdict(dict)
    results["mark"]["wrong_ns"] = set()
    domain_names = domains["mark"].keys()

    for domain, nameserver in domains["mark"].items():
        if nameserver != nameservers:
            logger.debug("%s has unmatching nameservers: %s", domain, nameserver)
            results["mark"]["wrong_ns"].add(domain)

    results["acme"]["missing"] = domain_names - domains["acme"]
    results["dnsrepo"]["missing"] = domain_names - domains["dnsrepo"]
    results["ncredir"]["missing"] = domains["acme"] - domains["ncredir"]

    results["acme"]["unused"] = domains["acme"] - domain_names
    results["dnsrepo"]["unused"] = domains["dnsrepo"] - domain_names
    results["ncredir"]["unused"] = domains["ncredir"] - domain_names

    logger.debug("Categorized results:\n%s", results)

    if pretty_print:
        results_json = json.dumps(results, sort_keys=True, cls=SetToListEncoder, indent=4)
    else:
        results_json = json.dumps(results, sort_keys=True, cls=SetToListEncoder)

    return results_json


def main() -> int:
    """
    Get acme-chief, ncredir, and MarkMonitor domain information. Compare and
    report on any inconsistencies.
    """
    args = parse_args()
    setup_logging(args)
    logger.debug(args)

    # Did the user supply a config file?
    if args.config:
        config = NCMonitorConfig(args.config)
    else:
        config = NCMonitorConfig(get_conf_file())

    domains = get_domains(config.mark_user, config.mark_pass)
    results_json = process_domains(config.nameservers, domains, args.pprint)
    logger.info("Final result of domains to address:")
    print(results_json)


if __name__ == '__main__':
    main()
